from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
    overload,
)

from dataclasses import Field, dataclass, field

from apischema.alias import ALIAS_METADATA

ErrorMsg = str
Error = Union[ErrorMsg, Tuple[Any, ErrorMsg]]


@dataclass
class LocalizedError:
    loc: Sequence[str]
    err: Union[ErrorMsg, Sequence[ErrorMsg]]

    def __post_init__(self):
        if not isinstance(self.err, str) and len(self.err) == 1:
            self.err = self.err[0]


@dataclass
class ValidationError(Exception):
    messages: Sequence[ErrorMsg] = field(default_factory=list)
    children: Mapping[str, "ValidationError"] = field(default_factory=dict)

    def flat(self) -> Iterator[Tuple[Tuple[str, ...], Sequence[ErrorMsg]]]:
        if self.messages:
            yield (), self.messages
        for child_path, child in self.children.items():
            for path, errors in child.flat():
                yield (child_path, *path), errors

    def format(self) -> Iterator[LocalizedError]:
        for loc, err in self.flat():
            yield LocalizedError(loc, err)


@overload
def merge(err1: Optional[ValidationError], err2: ValidationError) -> ValidationError:
    ...


@overload
def merge(err1: ValidationError, err2: Optional[ValidationError]) -> ValidationError:
    ...


@overload
def merge(
    err1: Optional[ValidationError], err2: Optional[ValidationError]
) -> Optional[ValidationError]:
    ...


def merge(err1: Optional[ValidationError], err2: Optional[ValidationError]):
    if err1 is None:
        return err2
    if err2 is None:
        return err1
    return ValidationError(
        [*err1.messages, *err2.messages],
        {
            key: merge(err1.children.get(key), err2.children.get(key))  # type: ignore
            for key in err1.children.keys() | err2.children.keys()
        },
    )


def exception(err: Exception) -> str:
    return f"[{type(err).__name__}]{err}"


def _rec_build_error(path: Sequence[str], msg: ErrorMsg) -> ValidationError:
    if not path:
        return ValidationError([msg])
    else:
        return ValidationError(children={path[0]: _rec_build_error(path[1:], msg)})


def _check_error_path(path: Any) -> Sequence[str]:
    if isinstance(path, (Field, int, str)):
        path = [path]
    else:
        path = list(path)
    for i, elt in enumerate(path):
        if isinstance(elt, Field):
            path[i] = elt.metadata.get(ALIAS_METADATA, elt.name)
        elif isinstance(elt, int):
            path[i] = str(elt)
        if not isinstance(path[i], str):
            raise TypeError(
                f"Bad error path, expected Field, int or str," f" found {type(i)}"
            )
    return path


def build_from_errors(errors: Iterable[Error]) -> ValidationError:
    messages: List[ErrorMsg] = []
    children: Dict[str, ValidationError] = {}
    for error in errors:
        if isinstance(error, ErrorMsg):
            messages.append(error)
            continue
        path, msg = error
        path = _check_error_path(path)
        if not path:
            messages.append(msg)
        else:
            children[path[0]] = merge(
                children.get(path[0]), _rec_build_error(path[1:], msg)
            )
    return ValidationError(messages, children)



from dataclasses import is_dataclass, replace
from enum import Enum
from itertools import chain
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Pattern,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)

from apischema.conversion import Converter, InputVisitorMixin, OutputVisitorMixin
from apischema.data import to_data
from apischema.dataclasses import (
    Field,
    FieldCache,
    FieldKind,
    get_input_fields,
    get_output_fields,
)
from apischema.fields import NoDefault, get_default, get_fields_set, mark_set_fields
from apischema.ignore import Ignored
from apischema.json_schema.types import JSONSchema, JSONType
from apischema.schema import Constraint, Schema
from apischema.schema.annotations import get_annotations
from apischema.schema.constraints import (
    ArrayConstraint,
    NumberConstraint,
    ObjectConstraint,
    StringConstraint,
    get_constraint,
)
from apischema.utils import NO_DEFAULT, as_dict
from apischema.visitor import Visitor

constraint_by_type = {
    int: NumberConstraint,
    float: NumberConstraint,
    str: StringConstraint,
}


class ReadWriteOnly(Constraint):
    pass


def check_constraint(schema: Schema, expected: Type[Constraint]):
    if schema.constraint is not None:
        if not isinstance(schema.constraint, expected):
            if not isinstance(schema.constraint, ReadWriteOnly):
                raise TypeError(
                    f"Bad constraint: expected {expected.__name__}"
                    f" found {type(schema.constraint).__name__}"
                )


def _remove_none(obj: Optional[Any]) -> Iterator[Tuple[str, Any]]:
    if obj is None:
        return
    for k, v in as_dict(obj).items():
        filtered = {"items", "additional_properties"}
        if v is not None and v is not NO_DEFAULT and k not in filtered:
            yield k, v


def _to_dict(schema: Schema) -> Mapping[str, Any]:
    return dict(
        chain(_remove_none(schema.annotations), _remove_none(schema.constraint))
    )


def _merge_constraints(
    c1: Optional[Constraint], c2: Optional[Constraint]
) -> Optional[Constraint]:
    if c1 is None:
        return c2
    if c2 is None:
        return c1
    if isinstance(c1, ReadWriteOnly):
        return type(c2)(**{**as_dict(c1), **as_dict(c2)})
    elif Constraint != type(c1) != type(c2) != Constraint:
        raise TypeError(
            f"Incompatibles constraints {type(c1).__name__}" f" and {type(c2).__name__}"
        )
    return c1


def _override(schema: Schema, cls: Type) -> Schema:
    return Schema(
        schema.annotations or get_annotations(cls),
        _merge_constraints(schema.constraint, get_constraint(cls)),
    )


def _field_schema(field: Field) -> Schema:
    annotations = field.annotations
    if annotations is not None and annotations.default is ...:
        try:
            default = to_data(get_default(field.base_field))
        except NoDefault:
            raise TypeError("Invalid ... without field default")
        annotations = replace(annotations, default=default)
    return Schema(annotations, field.constraint)


def _extract_properties_schema(json_schema: JSONSchema) -> JSONSchema:
    if json_schema.pattern_properties is not None:
        if (
            len(json_schema.pattern_properties) >= 1
            or json_schema.additional_properties
        ):  # don't try to merge the schemas and return
            return JSONSchema()
        return next(iter(json_schema.pattern_properties.values()))
    if json_schema.additional_properties:
        if isinstance(json_schema.additional_properties, JSONSchema):
            return json_schema.additional_properties
        else:
            return JSONSchema()
    raise TypeError("properties field must have an 'object' schema")


class SchemaBuilder(Visitor[Schema, JSONSchema]):
    def __init__(self, ref_factory: Optional[Callable[[Type], str]]):
        super().__init__()
        self.ref_factory = ref_factory or self._missing_ref_factory
        self.schemas: Set[Type] = set()

    @staticmethod
    def _missing_ref_factory(cls: Type) -> str:
        raise TypeError("reference factory is needed to handle" " recursive types")

    def primitive(self, cls: Type, schema: Schema) -> JSONSchema:
        if cls in constraint_by_type:
            check_constraint(schema, constraint_by_type[cls])
        return JSONSchema(type=JSONType.from_type(cls), **_to_dict(schema))

    def subprimitive(self, cls: Type, superclass: Type, schema: Schema) -> JSONSchema:
        return self.primitive(superclass, _override(schema, cls))

    def union(self, alternatives: Sequence[Type], schema: Schema) -> JSONSchema:
        any_of = []
        for cls in alternatives:
            try:
                any_of.append(self.visit(cls, schema))
            except Ignored:
                pass
        if len(any_of) == 1:
            return any_of[0]
        else:
            return JSONSchema(any_of=any_of)

    def iterable(
        self, cls: Type[Iterable], value_type: Type, schema: Schema
    ) -> JSONSchema:
        check_constraint(schema, ArrayConstraint)
        if isinstance(schema.items_, Sequence):
            raise TypeError("Iterable items schema cannot be a list of schema")
        return JSONSchema(
            type=JSONType.ARRAY,
            items=self.visit(value_type, schema.items_),
            **_to_dict(schema),
        )

    def mapping(
        self, cls: Type[Mapping], key_type: Type, value_type: Type, schema: Schema
    ) -> JSONSchema:
        check_constraint(schema, ObjectConstraint)
        key = self.visit(key_type, Schema())
        value = self.visit(value_type, schema.additional_properties)
        properties: Dict[str, Any] = {}
        if key.pattern is not None:
            properties["pattern_properties"] = {key.pattern: value}
        else:
            properties["additional_properties"] = value
        return JSONSchema(  # type: ignore
            type=JSONType.OBJECT, **properties, **_to_dict(schema),
        )

    def typed_dict(
        self, cls: Type, keys: Mapping[str, Type], total: bool, schema: Schema
    ) -> JSONSchema:
        if cls in self.schemas:
            return JSONSchema(ref=self.ref_factory(cls))
        check_constraint(schema, ObjectConstraint)
        if schema.additional_properties:
            raise TypeError("additional properties are not handled" " for TypedDict")
        sorted_keys = sorted(keys)
        return JSONSchema(
            type=JSONType.OBJECT,
            properties={key: self.visit(keys[key], Schema()) for key in sorted_keys},
            required=sorted_keys if total else [],
            additional_properties=True,
            **_to_dict(schema),
        )

    def tuple(self, types: Sequence[Type], schema: Schema) -> JSONSchema:
        check_constraint(schema, ArrayConstraint)
        if schema.constraint is not None:
            assert isinstance(schema.constraint, ArrayConstraint)
            if (
                schema.constraint.max_items is not None
                or schema.constraint.min_items is not None
            ):
                raise TypeError("Tuple cannot have min_items/max_items" " constraint")
        if isinstance(schema.items_, Sequence):
            if len(schema.items_) != len(types):
                raise TypeError(
                    "Tuple items schema list must have the same" " length than tuple"
                )
        items = []
        for i, cls in enumerate(types):
            if isinstance(schema.items_, Sequence):
                elt_schema = schema.items_[i]
            else:
                elt_schema = schema.items_
            items.append(self.visit(cls, elt_schema))
        return JSONSchema(
            type=JSONType.ARRAY,
            items=items,
            min_items=len(types),
            max_items=len(types),
            **_to_dict(schema),
        )

    def literal(self, values: Sequence[Any], schema: Schema) -> JSONSchema:
        if not values:
            raise TypeError("empty Literal")
        if schema.constraint is not None:
            raise TypeError("Literal cannot have constraint")
        types = sorted(
            set(
                JSONType.from_type(type(v.value if isinstance(v, Enum) else v))
                for v in values
            )
        )
        # Mypy issue
        type_: Any = types[0] if len(types) == 1 else types
        if len(values) == 1:
            return JSONSchema(type=type_, const=values[0], **_to_dict(schema),)
        else:
            return JSONSchema(type=type_, enum=values, **_to_dict(schema),)

    def _dataclass_fields(self, cls: Type) -> FieldCache:
        raise NotImplementedError()

    def _field_visit(self, field: Field) -> JSONSchema:
        raise NotImplementedError()

    def dataclass(self, cls: Type, schema: Schema) -> JSONSchema:
        assert is_dataclass(cls)
        if cls in self.schemas:
            return JSONSchema(ref=self.ref_factory(cls))
        self.schemas.add(cls)
        check_constraint(schema, ObjectConstraint)
        schema_ = _override(schema, cls)
        properties = {}
        required: List[str] = []
        fields, pattern_fields, additional_field = self._dataclass_fields(cls)
        for field in fields:
            properties[field.alias] = self._field_visit(field)
            if not field.default:
                required.append(field.alias)
        pattern_properties: Dict[Pattern, JSONSchema] = {}
        for pattern, field in pattern_fields:
            field_schema = self._field_visit(field)
            pattern_properties[pattern] = _extract_properties_schema(field_schema)
        additional_properties: Optional[Union[bool, "JSONSchema"]] = False
        if additional_field:
            field_schema = self._field_visit(additional_field)
            additional_properties = _extract_properties_schema(field_schema)
        kwargs: Dict[str, Any] = {}
        if properties:
            kwargs["properties"] = properties
        if required:
            kwargs["required"] = required
        # get_fields_set(additional_properties) means that additional_properties
        # is not {} and not True
        if additional_properties is False or get_fields_set(additional_properties):
            kwargs["additional_properties"] = additional_properties
        if pattern_properties:
            kwargs["pattern_properties"] = pattern_properties
        return JSONSchema(  # type: ignore
            type=JSONType.OBJECT, **kwargs, **_to_dict(schema_),
        )

    def enum(self, cls: Type[Enum], schema: Schema) -> JSONSchema:
        if schema.constraint is not None:
            raise TypeError("enum cannot have constraint")
        if len(cls) == 0:
            raise TypeError("Empty enum")
        return self.literal(list(cls), _override(schema, cls))

    def new_type(self, cls: Type, super_type: Type, schema: Schema) -> JSONSchema:
        return self.visit(super_type, _override(schema, cls))

    def any(self, schema: Schema) -> JSONSchema:
        if schema.constraint is not None:
            raise TypeError("Any type cannot have constraint")
        return JSONSchema()

    def annotated(
        self, cls: Type, annotations: Sequence[Any], schema: Schema
    ) -> JSONSchema:
        schema_ = schema
        if Ignored in annotations:
            raise Ignored
        for annotation in annotations:
            if isinstance(annotation, Schema):
                schema_ = Schema(
                    schema.annotations or annotation.annotations,
                    _merge_constraints(schema.constraint, annotation.constraint),
                )
                break
        return self.visit(cls, schema_)


class InputSchemaBuilder(InputVisitorMixin[Schema, JSONSchema], SchemaBuilder):
    def __init__(
        self, ref_factory: Optional[Callable[[Type], str]], additional_properties: bool
    ):
        SchemaBuilder.__init__(self, ref_factory)
        self.additional_properties = additional_properties

    def _custom(
        self, cls: Type, custom: Dict[Type, Converter], schema: Schema
    ) -> JSONSchema:
        return self.union(list(custom), _override(schema, cls))

    _dataclass_fields = staticmethod(get_input_fields)  # type: ignore

    def _field_visit(self, field: Field) -> JSONSchema:
        schema = _field_schema(field)
        if field.kind == FieldKind.INIT:
            if schema.constraint is None:
                schema.constraint = ReadWriteOnly(write_only=True)
            else:
                if schema.constraint.read_only:
                    raise TypeError("InitVar cannot be read-only")
                schema.constraint = replace(schema.constraint, write_only=True)
        return self.visit(field.input_type, schema)

    def dataclass(self, cls: Type, schema: Schema) -> JSONSchema:
        json_schema = super().dataclass(cls, schema)
        if json_schema.additional_properties is False:
            # TODO document that `dataclasses.replace` don't keep fields set
            return mark_set_fields(
                replace(json_schema, additional_properties=self.additional_properties),
                *get_fields_set(json_schema),
                overwrite=True,
            )
        return json_schema


def build_input_schema(
    cls: Type,
    *,
    ref_factory: Callable[[Type], str] = None,
    additional_properties: bool = False,
    schema: Schema = None,
) -> JSONSchema:
    builder = InputSchemaBuilder(ref_factory, additional_properties)
    return builder.visit(cls, schema or Schema())


class OutputSchemaBuilder(OutputVisitorMixin[Schema, JSONSchema], SchemaBuilder):
    def __init__(
        self,
        conversions: Mapping[Type, Type],
        ref_factory: Optional[Callable[[Type], str]],
    ):
        SchemaBuilder.__init__(self, ref_factory)
        OutputVisitorMixin.__init__(self, conversions)

    def _custom(
        self, cls: Type, custom: Tuple[Type, Converter], schema: Schema
    ) -> JSONSchema:
        return self.visit(custom[0], _override(schema, cls))

    _dataclass_fields = staticmethod(get_output_fields)  # type: ignore

    def _field_visit(self, field: Field) -> JSONSchema:
        schema = _field_schema(field)
        if field.kind == FieldKind.NO_INIT:
            if schema.constraint is None:
                schema.constraint = ReadWriteOnly(read_only=True)
            else:
                if schema.constraint.write_only:
                    raise TypeError("not init field cannot be write-only")
                schema.constraint = replace(schema.constraint, read_only=True)
        return self.visit(field.output_type, schema)


def build_output_schema(
    cls: Type,
    conversions: Mapping[Type, Type] = None,
    *,
    ref_factory: Callable[[Type], str] = None,
    schema: Schema = None,
) -> JSONSchema:
    builder = OutputSchemaBuilder(conversions or {}, ref_factory)
    return builder.visit(cls, schema or Schema())



"""Kind of typing_extensions for this package"""
__all__ = [
    "Annotated",
    "Literal",
    "NamedTupleMeta",
    "Protocol",
    "TypedDict",
    "_AnnotatedAlias",
    "_GenericAlias",
    "_LiteralMeta",
    "_TypedDictMeta",
    "get_type_hints",
    "set_type_hints",
    "_type_repr",
]

import sys
import types
from typing import Any, Dict, Generic, Mapping, NamedTuple, Optional, Type, TypeVar

from dataclasses import is_dataclass


class NoType:
    def __getitem__(self, key):
        return self


NO_TYPE: Any = NoType()

if sys.version_info >= (3, 9):
    from typing import Annotated, get_type_hints as gth
else:
    try:
        from typing_extensions import Annotated
    except ImportError:
        Annotated = NO_TYPE
    try:
        from typing_extensions import get_type_hints as gth
    except ImportError:
        from typing import get_type_hints as gth_

        def gth(obj, globalns=None, localns=None, include_extras=False):  # type: ignore
            return gth_(obj, globalns, localns)


if sys.version_info >= (3, 8):
    from typing import Literal, Protocol, TypedDict
else:
    try:
        from typing_extensions import Literal, Protocol, TypedDict
    except ImportError:
        Literal = NO_TYPE
        TypedDict = NO_TYPE
        Protocol = NO_TYPE

Annotated = Annotated
Literal = Literal
Protocol = Protocol
TypedDict = TypedDict

_T = TypeVar("_T")
_AnnotatedAlias = type(Annotated[_T, ...])
_GenericAlias = type(Generic[_T])
_LiteralMeta = type(Literal)
_TypedDictMeta = type(TypedDict)
NamedTupleMeta = type(NamedTuple)

_type_hints: Dict[str, Mapping[str, Type]] = {}

NS = Dict[str, Any]


def get_type_hints(
    obj,
    globalns: Optional[NS] = None,
    localns: Optional[NS] = None,
    include_extras=False,
):
    try:
        return _type_hints[obj]
    except (KeyError, TypeError):
        types = gth(obj, globalns, localns, include_extras=True)
        if isinstance(obj, _TypedDictMeta) or is_dataclass(NamedTupleMeta):
            try:
                _type_hints[obj] = types
            except TypeError:
                pass
        return types


set_type_hints = get_type_hints

try:  # pragma: no cover
    from typing import _type_repr  # type:ignore
except ImportError:
    # copy from typing._type_repr in case of ...
    def _type_repr(obj):
        if isinstance(obj, type):
            if obj.__module__ == "builtins":
                return obj.__qualname__
            return f"{obj.__module__}.{obj.__qualname__}"
        if obj is ...:
            return "..."
        if isinstance(obj, types.FunctionType):
            return obj.__name__
        return repr(obj)



from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Generic,
    Hashable,
    Iterable,
    TypeVar,
    Union,
)

from dataclasses import fields, is_dataclass

PREFIX = "_apischema_"
NO_DEFAULT = object()

Func = TypeVar("Func", bound=Callable)

Hashable_ = TypeVar("Hashable_", bound=Hashable)


def to_hashable(data: Union[None, int, float, str, bool, list, dict]) -> Hashable:
    if isinstance(data, list):
        return tuple(map(to_hashable, data))
    if isinstance(data, dict):
        return tuple(sorted((to_hashable(k), to_hashable(v)) for k, v in data.items()))
    return data  # type: ignore


T = TypeVar("T")
U = TypeVar("U")


class GeneratorValue(Iterable[U], Generic[T, U]):
    value: T

    def __init__(self, generator: Generator[U, None, T]):
        self.generator = generator

    def __iter__(self):
        self.value = yield from self.generator


def to_camel_case(s: str):
    pascal_case = "".join(map(str.capitalize, s.split("_")))
    return pascal_case[0].lower() + pascal_case[1:]


def as_dict(obj) -> Dict[str, Any]:
    assert is_dataclass(obj)
    return {f.name: getattr(obj, f.name) for f in fields(obj)}



from functools import partial
from types import FunctionType, MethodType
from typing import Any, Callable, Mapping, Optional, Type

from dataclasses import Field, _FIELDS, _FIELD_CLASSVAR, dataclass  # type: ignore

from apischema.fields import FIELDS_SET_ATTR, get_default

MOCK_FIELDS_FIELD = "__mock_fields__"
MOCK_CLS_FIELD = "__mock_cls__"


class NonTrivialDependency(Exception):
    def __init__(self, attr: str):
        self.attr = attr
        self.validator: Optional[Callable] = None


@dataclass
class ValidatorMock:
    def __init__(
        self, cls: Type, fields: Mapping[str, Any], defaults: Mapping[str, Field]
    ):
        self.cls = cls
        self.fields = fields
        self.defaults = defaults
        setattr(self, FIELDS_SET_ATTR, set(fields))

    def __getattribute__(self, attr: str) -> Any:
        fields = super().__getattribute__("fields")
        if attr in fields:
            return fields[attr]
        defaults = super().__getattribute__("defaults")
        if attr in defaults:
            return get_default(defaults[attr])
        cls = super().__getattribute__("cls")
        if attr == "__class__":
            return cls
        if attr == "__dict__":
            return {
                **fields,
                **{name: get_default(field) for name, field in defaults.items()},
                FIELDS_SET_ATTR: set(fields),
            }
        if hasattr(cls, attr):
            member = getattr(cls, attr)
            # for classmethod (staticmethod are not handled)
            if isinstance(member, MethodType):
                return member
            if isinstance(member, FunctionType):
                return partial(member, self)
            if isinstance(member, property):
                return member.fget(self)  # type: ignore
            if all(
                f.name != attr or f._field_type == _FIELD_CLASSVAR
                for f in getattr(cls, _FIELDS).values()
            ):
                return member
        raise NonTrivialDependency(attr)



__all__ = ["get_schema", "schema", "Schema", "Annotations", "Constraint"]

from typing import Any, Dict, Optional, Sequence, Type, TypeVar, Union, overload

from dataclasses import dataclass, fields

from apischema.types import MetadataMixin, Number
from .annotations import (
    ANNOTATIONS_METADATA,
    Annotations,
    _annotations,
    get_annotations,
)
from .constraints import (
    ArrayConstraint,
    CONSTRAINT_METADATA,
    Constraint,
    NumberConstraint,
    ObjectConstraint,
    StringConstraint,
    _constraints,
    get_constraint,
)

T = TypeVar("T")


@dataclass
class Schema(MetadataMixin):
    annotations: Optional[Annotations] = None
    constraint: Optional[Constraint] = None

    def __post_init__(self):
        self.metadata: Dict[str, Any] = {}
        if self.annotations is not None:
            self.metadata[ANNOTATIONS_METADATA] = self.annotations
        if self.constraint is not None:
            self.metadata[CONSTRAINT_METADATA] = self.constraint

    def validate(self, _obj: T) -> T:
        if self.constraint is not None:
            self.constraint.validate(_obj)
        return _obj

    def __call__(self, obj: T) -> T:
        if self.annotations is not None:
            _annotations[obj] = self.annotations
        if self.constraint is not None:
            _constraints[obj] = self.constraint
        return obj

    @property
    def items_(self) -> Union["Schema", Sequence["Schema"]]:
        assert self.constraint is None or isinstance(self.constraint, ArrayConstraint)
        annotations, constraint = None, None
        if self.annotations is not None:
            annotations = self.annotations.items
        if self.constraint is not None:
            constraint = self.constraint.items
        if isinstance(annotations, Sequence):
            assert isinstance(constraint, Sequence)
            assert len(annotations) == len(constraint)
            return [Schema(a, c) for a, c in zip(annotations, constraint)]
        else:
            assert annotations is None or isinstance(annotations, Annotations)
            assert constraint is None or isinstance(constraint, Constraint)
            return Schema(annotations, constraint)

    @property
    def additional_properties(self) -> "Schema":
        assert self.constraint is None or isinstance(self.constraint, ObjectConstraint)
        annotations, constraint = None, None
        if self.annotations is not None:
            annotations = self.annotations.additional_properties
        if self.constraint is not None:
            constraint = self.constraint.additional_properties
        return Schema(annotations, constraint)


def get_schema(cls: Type) -> Schema:
    return Schema(get_annotations(cls), get_constraint(cls))


_constraint_rewrite = {
    "min": "minimum",
    "max": "maximum",
    "exc_min": "exclusive_minimum",
    "exc_max": "exclusive_maximum",
    "mult_of": "multiple_of",
    "min_len": "min_length",
    "max_len": "max_length",
    "unique": "unique_items",
    "properties": "additional_properties",
}


@overload
def schema(
    *,
    title: Optional[str] = None,
    description: Optional[str] = None,
    examples: Optional[Sequence[Any]] = None,
    read_only: Optional[bool] = None,
    write_only: Optional[bool] = None
) -> Schema:
    ...


@overload
def schema(
    *,
    title: Optional[str] = None,
    description: Optional[str] = None,
    examples: Optional[Sequence[Any]] = None,
    read_only: Optional[bool] = None,
    write_only: Optional[bool] = None,
    min: Optional[Number] = None,
    max: Optional[Number] = None,
    exc_min: Optional[Number] = None,
    exc_max: Optional[Number] = None,
    mult_of: Optional[Number] = None
) -> Schema:
    ...


@overload
def schema(
    *,
    title: Optional[str] = None,
    description: Optional[str] = None,
    examples: Optional[Sequence[Any]] = None,
    read_only: Optional[bool] = None,
    write_only: Optional[bool] = None,
    format: Optional[str] = None,
    min_len: Optional[int] = None,
    max_len: Optional[int] = None,
    pattern: Optional[str] = None
) -> Schema:
    ...


@overload
def schema(
    *,
    title: Optional[str] = None,
    description: Optional[str] = None,
    examples: Optional[Sequence[Any]] = None,
    read_only: Optional[bool] = None,
    write_only: Optional[bool] = None,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique: Optional[bool] = None,
    items: Optional[Union[Schema, Sequence[Schema]]] = None
) -> Schema:
    ...


@overload
def schema(
    *,
    title: Optional[str] = None,
    description: Optional[str] = None,
    examples: Optional[Sequence[Any]] = None,
    read_only: Optional[bool] = None,
    write_only: Optional[bool] = None,
    min_properties: Optional[int] = None,
    max_properties: Optional[int] = None,
    properties: Optional[Schema] = None
) -> Schema:
    ...


def schema(**kwargs) -> Schema:
    kwargs_ = {_constraint_rewrite.get(k, k): v for k, v in kwargs.items()}
    annotations_fields = {f.name for f in fields(Annotations)}
    annotations_kwargs = {k: v for k, v in kwargs_.items() if k in annotations_fields}
    constraint_kwargs = {
        k: v for k, v in kwargs_.items() if k not in annotations_kwargs
    }
    if "additional_properties" in annotations_kwargs:
        schema1: Schema = annotations_kwargs["additional_properties"]
        annotations_kwargs["additional_properties"] = schema1.annotations
        constraint_kwargs["additional_properties"] = schema1.constraint
    if "items" in annotations_kwargs:
        schema2: Union[Schema, Sequence[Schema]] = annotations_kwargs["items"]
        if not isinstance(schema2, Sequence):
            annotations_kwargs["items"] = [s.annotations for s in schema2]
            constraint_kwargs["items"] = [s.constraint for s in schema2]
        else:
            assert isinstance(schema2, Schema)
            annotations_kwargs["items"] = schema2.annotations
            constraint_kwargs["items"] = schema2.constraint

    annotations = None
    if annotations_kwargs:
        annotations = Annotations(**annotations_kwargs)

    constraint = None
    if constraint_kwargs:
        for cls in (
            NumberConstraint,
            StringConstraint,
            ArrayConstraint,
            ObjectConstraint,
        ):
            try:
                constraint = cls(**constraint_kwargs)
                break
            except TypeError:
                pass

    return Schema(annotations, constraint)



import collections.abc
import sys
from typing import (
    AbstractSet,
    Any,
    Collection,
    Dict,
    FrozenSet,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    MutableSequence,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)

from dataclasses import Field

AnyType = Any
NoneType: Type[None] = type(None)
Number = Union[int, float]

PRIMITIVE_TYPES = {str, int, bool, float, NoneType}

if sys.version_info >= (3, 7):
    ITERABLE_TYPES = {
        collections.abc.Iterable: tuple,
        collections.abc.Collection: tuple,
        collections.abc.Sequence: tuple,
        tuple: tuple,
        collections.abc.MutableSequence: list,
        list: list,
        collections.abc.Set: frozenset,
        frozenset: frozenset,
        collections.abc.MutableSet: set,
        set: set,
    }

    LIST_TYPE = list

    MAPPING_TYPES = {
        collections.abc.Mapping: dict,
        collections.abc.MutableMapping: dict,
        dict: dict,
    }
else:
    ITERABLE_TYPES = {
        Iterable: tuple,
        Collection: tuple,
        Sequence: tuple,
        Tuple: tuple,
        MutableSequence: list,
        List: list,
        AbstractSet: frozenset,
        FrozenSet: frozenset,
        Set: set,
    }

    LIST_TYPE = list

    MAPPING_TYPES = {Mapping: dict, MutableMapping: dict, Dict: dict}


if sys.version_info >= (3, 7):
    OrderedDict = dict
else:
    from collections import OrderedDict  # noqa

# Kind of hack to benefit of PEP 584
if sys.version_info >= (3, 9):
    Metadata = Mapping[str, Any]
    DictWithUnion = dict
else:

    class Metadata(Mapping[str, Any]):
        def __or__(self, other: Mapping[str, Any]) -> "Metadata":
            return DictWithUnion({**self, **other})

    class DictWithUnion(Dict[str, Any], Metadata):  # noqa
        pass


class MetadataMixin(Metadata):
    metadata: Mapping[str, Any]

    def __getitem__(self, key):
        return self.metadata[key]

    def __iter__(self):
        return iter(self.metadata)

    def __len__(self):
        return len(self.metadata)

    def _resolve(self, field: Field, field_type: AnyType) -> Metadata:
        return self



import sys
from collections import defaultdict
from inspect import Parameter, isclass, isgeneratorfunction, signature
from logging import getLogger
from types import MethodType
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Generic,
    Mapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from dataclasses import MISSING, dataclass, field

from apischema.properties import properties
from apischema.types import DictWithUnion, Metadata
from apischema.typing import _GenericAlias, get_type_hints
from apischema.utils import GeneratorValue, PREFIX, as_dict, to_camel_case
from apischema.validation.errors import build_from_errors
from apischema.visitor import NOT_CUSTOM, NotCustom

logger = getLogger(__name__)

Converter = Callable[[Any], Any]

_input_converters: Dict[Type, Dict[Type, Converter]] = defaultdict(dict)
_output_converters: Dict[Type, Tuple[Type, Converter]] = {}
_converters: Dict[Type, Dict[Type, Converter]] = defaultdict(dict)

Param = TypeVar("Param")
Return = TypeVar("Return")


def check_converter(
    converter: Converter, param: Optional[Type[Param]], ret: Optional[Type[Return]]
) -> Tuple[Type[Param], Type[Return]]:
    try:
        parameters = iter(signature(converter).parameters.values())
    except ValueError:  # builtin types
        if ret is None and isclass(ret):
            ret = cast(Type[Any], converter)
        if param is None:
            raise TypeError("converter parameter must be typed")
    else:
        try:
            first = next(parameters)
        except StopIteration:
            raise TypeError("converter must have at least one parameter")
        for p in parameters:
            if p.default is Parameter.empty and p.kind not in (
                Parameter.VAR_POSITIONAL,
                Parameter.VAR_KEYWORD,
            ):
                raise TypeError(
                    "converter must have at most one parameter " "without default"
                )
        types = get_type_hints(converter, include_extras=True)
        if param is None:
            try:
                param = types.pop(first.name)
            except KeyError:
                raise TypeError("converter parameter must be typed")
        if ret is None:
            try:
                ret = types.pop("return")
            except KeyError:
                raise TypeError("converter return must be typed")
    return cast(Type[Param], param), cast(Type[Return], ret)


def handle_potential_validation(
    ret: Type, converter: Converter
) -> Tuple[Type, Converter]:
    if not isgeneratorfunction(converter):
        return ret, converter

    def wrapper(arg):
        generator = GeneratorValue(converter(arg))
        errors = [*generator]
        if errors:
            raise build_from_errors(errors)
        return generator.value

    if getattr(ret, "__origin__", None) == Generator:
        return ret.__args__[2], wrapper
    else:
        return ret, wrapper


Func = TypeVar("Func", bound=Callable)


class MethodConverter:
    def __init__(self, decorator: Callable, func: Callable):
        self.decorator = decorator
        self.func = func

    def __get__(self, instance, owner):
        return self if instance is None else MethodType(self.func, instance)

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __set_name__(self, owner, name):
        self.decorator(self.func, owner)


def handle_method(decorator: Func) -> Func:
    def decorator2(func, param=None, ret=None):
        if param is None and ret is None:
            if list(signature(func).parameters)[0] == "self":
                return MethodConverter(decorator, func)
        return decorator(func, param, ret)

    return cast(Func, decorator2)


Cls = TypeVar("Cls", bound=Type)
Other = TypeVar("Other", bound=Type)

if sys.version_info >= (3, 7):
    import collections.abc
    import re
    from typing import (
        List,
        AbstractSet,
        Set,
        Dict,
        Iterable,
        Collection,
        Sequence,
        MutableSequence,
        Pattern,
    )

    TYPED_ORIGINS = {
        tuple: Tuple,
        list: List,
        frozenset: AbstractSet,
        set: Set,
        dict: Dict,
        collections.abc.Iterable: Iterable,
        collections.abc.Collection: Collection,
        collections.abc.Sequence: Sequence,
        collections.abc.MutableSequence: MutableSequence,
        collections.abc.Set: AbstractSet,
        collections.abc.MutableSet: Set,
        re.Pattern: Pattern,
    }

    def _get_origin(cls: _GenericAlias) -> Type:  # type: ignore
        return TYPED_ORIGINS.get(cls.__origin__, cls.__origin__)  # type: ignore


else:

    def _get_origin(cls: _GenericAlias) -> Type:  # type: ignore
        return cls.__origin__  # type: ignore


# If A[str, T] and A[T, int], what to choose for A[str, int]? None of them.


def substitute_generic_args(cls: Cls, substitution: Mapping) -> Cls:
    if getattr(cls, "__origin__", None) is None:
        return cls
    args = tuple(substitution.get(arg, arg) for arg in cls.__args__)
    return _get_origin(cls)[args]


def substitute_type_vars(base: Cls, other: Other) -> Tuple[Cls, Other]:
    if getattr(base, "__origin__", None) is None:
        return base, other
    substitution = {
        arg: param
        for arg, param in zip(base.__args__, _get_origin(base).__parameters__)
        if isinstance(arg, TypeVar)  # type: ignore
    }
    if isinstance(other, TypeVar):  # type: ignore
        new_other = substitution.get(other, other)
    else:
        new_other = substitute_generic_args(other, substitution)
    if all(isinstance(arg, TypeVar) for arg in base.__args__):  # type: ignore
        return cast(Tuple[Cls, Other], (_get_origin(base), new_other))
    else:
        return cast(
            Tuple[Cls, Other], (substitute_generic_args(base, substitution), new_other)
        )


@handle_method
def converter(function: Func, param: Type = None, ret: Type = None) -> Func:
    param, ret = check_converter(function, param, ret)
    param, ret = substitute_type_vars(param, ret)
    if hasattr(param, "__supertype__"):
        # For performance reason, serialization is based on objects class
        raise TypeError("NewType cannot be used with converter/output_converter")
    _converters[param][ret] = function
    return function


@handle_method
def input_converter(function: Func, param: Type = None, ret: Type = None) -> Func:
    param, ret = check_converter(function, param, ret)
    ret, function_ = handle_potential_validation(ret, function)
    ret, param = substitute_type_vars(ret, param)
    _input_converters[ret][param] = function_
    return function


@handle_method
def output_converter(function: Func, param: Type = None, ret: Type = None) -> Func:
    param, ret = check_converter(function, param, ret)
    param, ret = substitute_type_vars(param, ret)
    _output_converters[param] = ret, function
    _converters[param][ret] = function
    return function


def inout_str(cls: Cls) -> Cls:
    input_converter(cls, str, cls)
    output_converter(str, cls, str)
    return cls


Arg = TypeVar("Arg")


class InputVisitorMixin(Generic[Arg, Return]):
    def _custom(self, cls: Type, custom: Dict[Type, Converter], arg: Arg) -> Return:
        raise NotImplementedError()

    def custom(self, cls: Type, arg: Arg) -> Union[Return, NotCustom]:
        if cls not in _input_converters:
            return NOT_CUSTOM
        return self._custom(cls, _input_converters[cls], arg)


class OutputVisitorMixin(Generic[Arg, Return]):
    def __init__(self, conversions: Mapping[Type, Type]):
        self.converters = {}
        for cls, out in conversions.items():
            cls, out = substitute_type_vars(cls, out)
            if out not in _converters.get(cls, {}):
                logger.warning(f"{cls} cannot be converted to {out}")
            else:
                self.converters[cls] = out, _converters[cls][out]

    def _custom(self, cls: Type, custom: Tuple[Type, Converter], arg: Arg) -> Return:
        raise NotImplementedError()

    def custom(self, cls: Type, arg: Arg) -> Union[Return, NotCustom]:
        if cls in self.converters:
            return self._custom(cls, self.converters[cls], arg)
        if cls in _output_converters:
            return self._custom(cls, _output_converters[cls], arg)
        return NOT_CUSTOM


INPUT_METADATA = f"{PREFIX}input"
OUTPUT_METADATA = f"{PREFIX}output"


def field_input_converter(
    converter: Callable[[Any], Any], param: Type = None, ret: Type = None
) -> Metadata:
    try:
        param, ret = check_converter(converter, param, ret)
    except NameError:
        pass
    else:
        ret, converter = handle_potential_validation(ret, converter)
    return DictWithUnion({INPUT_METADATA: (param, converter)})


def field_output_converter(
    converter: Callable[[Any], Any], param: Type = None, ret: Type = None
) -> Metadata:
    try:
        param, ret = check_converter(converter, param, ret)
    except NameError:
        pass
    return DictWithUnion({OUTPUT_METADATA: (ret, converter)})


def converter_from_raw(func: Callable) -> Converter:
    types = get_type_hints(func, include_extras=True)
    sig = signature(func)
    annotations: Dict[str, Any] = {}
    fields: Dict[str, Any] = {}
    kwargs = None
    for name, param in sig.parameters.items():
        if param.kind == Parameter.POSITIONAL_ONLY:
            raise TypeError("Forbidden positional-only parameter")
        if param.kind == Parameter.VAR_POSITIONAL:
            raise TypeError("Forbidden variadic positional parameter")
        if param.kind == Parameter.VAR_KEYWORD:
            fields[name] = field(default_factory=dict, metadata=properties(),)
            type_ = types.get(name, Any)
            annotations[name] = Mapping[str, type_]  # type: ignore
            kwargs = name
            continue
        try:
            default = MISSING
            if param.default != Parameter.empty:
                default = param.default
            fields[name] = default
            annotations[name] = types[name]
        except KeyError:
            raise TypeError("All parameters must be annotated")

    def converter(obj):
        kw = as_dict(obj)
        kw.update(kw.pop(kwargs, ()))
        return func(**kw)

    namespace = {**fields, "__annotations__": annotations}  # type: ignore
    cls: Type = dataclass(type(to_camel_case(func.__name__), (), namespace))
    converter.__annotations__ = {"obj": cls, "return": types.get("return")}
    return converter


def raw_input_converter(func: Func) -> Func:
    input_converter(converter_from_raw(func))
    return func


def raw_field_input_converter(func: Callable) -> Metadata:
    return field_input_converter(raw_input_converter(func))



from apischema.types import DictWithUnion, Metadata
from apischema.utils import PREFIX


class Ignored(Exception):
    pass


IGNORE_METADATA = f"{PREFIX}ignore"


def ignore_default() -> Metadata:
    return DictWithUnion({IGNORE_METADATA: True})



import re
import sys
from datetime import date, datetime, time
from ipaddress import (
    IPv4Address,
    IPv4Interface,
    IPv4Network,
    IPv6Address,
    IPv6Interface,
    IPv6Network,
)
from uuid import UUID

from apischema.conversion import inout_str, input_converter, output_converter
from apischema.schema import schema

# ==================== uuid ====================

inout_str(UUID)
schema(format="uuid")(UUID)

# ==================== datetime ====================

if sys.version_info >= (3, 7):
    input_converter(date.fromisoformat, str, date)
    output_converter(date.isoformat, date, str)
    schema(format="date")(date)

    input_converter(datetime.fromisoformat, str, datetime)
    output_converter(datetime.isoformat, datetime, str)
    schema(format="date-time")(datetime)

    input_converter(time.fromisoformat, str, time)
    input_converter(time.isoformat, time, str)
    schema(format="time")(time)

# ==================== ipaddress ====================

inout_str(IPv4Address)
schema(format="ipv4")(IPv4Address)

inout_str(IPv4Interface)
schema(format="ipv4")(IPv4Interface)

inout_str(IPv4Network)
schema(format="ipv4")(IPv4Network)

inout_str(IPv6Address)
schema(format="ipv6")(IPv6Address)

inout_str(IPv6Interface)
schema(format="ipv6")(IPv6Interface)

inout_str(IPv6Network)
schema(format="ipv6")(IPv6Network)

# ==================== pattern ====================
Pattern = type(re.compile(r""))
input_converter(re.compile, str, Pattern)
output_converter(lambda p: p.pattern, Pattern, str)



from functools import wraps
from inspect import Parameter, isfunction, isgeneratorfunction, signature
from types import MethodType
from typing import (
    AbstractSet,
    Any,
    Callable,
    Iterable,
    Optional,
    Sequence,
    TypeVar,
    overload,
)

from dataclasses import Field

from apischema.types import AnyType, DictWithUnion, Metadata
from apischema.typing import NO_TYPE, Protocol
from apischema.utils import PREFIX
from apischema.validation.dependencies import find_all_dependencies
from apischema.validation.errors import (
    ValidationError,
    build_from_errors,
    exception,
    merge,
)
from apischema.validation.mock import NonTrivialDependency

if Protocol is not NO_TYPE:

    class ValidatorFunc(Protocol):
        @overload
        def __call__(self, _obj):
            ...

        @overload
        def __call__(self, _obj, **kwargs):
            ...

        def __call__(self, _obj, **kwargs):
            ...

    class BaseValidator(Protocol):
        @overload
        def validate(self, _obj):
            ...

        @overload
        def validate(self, _obj, **kwargs):
            ...

        def validate(self, _obj, **kwargs):
            ...


else:
    ValidatorFunc = Callable  # type: ignore
    BaseValidator = Callable  # type: ignore

# use attribute instead of global dict in order to be inherited
VALIDATORS_ATTR = f"{PREFIX}validators"


class SimpleValidator:
    def __init__(self, func: "ValidatorFunc"):
        self.func = func
        parameters = signature(func).parameters
        validate = func
        if not any(
            param.kind == Parameter.VAR_KEYWORD for param in parameters.values()
        ):
            if not parameters:
                raise TypeError("validator must have at least one parameter")
            params = set(list(parameters)[1:])
            wrapped = validate

            def validate(_obj, **kwargs):
                kwargs2 = {k: v for k, v in kwargs.items() if k in params}
                return wrapped(_obj, **kwargs2)

        if isgeneratorfunction(func):
            wrapped = validate

            def validate(_obj, **kwargs):
                errors = [*validate(_obj, **kwargs)]
                if errors:
                    raise build_from_errors(errors)

        self.validate = validate  # type: ignore

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def validate(self, _obj, **kwargs):  # for typing
        ...


class Validator(SimpleValidator):
    def __init__(self, func: Callable):
        if isgeneratorfunction(func):

            @wraps(func)
            def validate(_obj, **kwargs):
                errors = []
                try:
                    errors.extend(func(_obj, **kwargs))
                except ValidationError:
                    raise RuntimeError("Validation error raised in generator validator")
                except Discard as err:
                    err.error = build_from_errors(errors)
                    raise
                if errors:
                    raise build_from_errors(errors)

            validate.__annotations__.pop("return", ...)
            super().__init__(validate)
        else:
            super().__init__(func)
        wraps(func)(self)
        self.func = func

    def __get__(self, instance, owner):
        return self if instance is None else MethodType(self.func, instance)

    def __set_name__(self, owner, name: str):
        self.dependencies = find_all_dependencies(owner, self.func)
        setattr(owner, VALIDATORS_ATTR, (*get_validators(owner), self))

    def can_be_called(self, fields: AbstractSet[str]) -> bool:
        return all(dep in fields for dep in self.dependencies)


Func = TypeVar("Func", bound=Callable)


def get_validators(obj) -> Sequence[Validator]:
    return getattr(obj, VALIDATORS_ATTR, ())


def add_validator(cls: AnyType, *validators: ValidatorFunc):
    for func in validators:
        validator = SimpleValidator(func)
        setattr(cls, VALIDATORS_ATTR, (*get_validators(cls), validator))


class Discard(Exception):
    def __init__(self, *fields):
        if not all(isinstance(f, Field) for f in fields):
            raise TypeError("Only fields can be discarded")
        self.fields: AbstractSet[str] = {f.name for f in fields}
        self.error: Optional[ValidationError] = None


T = TypeVar("T")


def validate(_obj: T, _validators: Sequence[BaseValidator] = None, **kwargs) -> T:
    if _validators is None:
        _validators = get_validators(_obj)
    if not _validators:
        return _obj
    error: Optional[ValidationError] = None
    while True:
        for i, validator in enumerate(_validators):
            try:
                validator.validate(_obj, **kwargs)
            except ValidationError as err:
                error = merge(error, err)
            except Discard as err:
                if err.error is None:
                    raise RuntimeError("Discard can only be raised in class Validator")
                error = merge(error, err.error)
                _validators = [
                    v
                    for v in _validators[i + 1 :]
                    if not isinstance(v, Validator) or not v.dependencies & err.fields
                ]
                break
            except NonTrivialDependency as exc:
                assert isinstance(validator, Validator)
                exc.validator = validator
                raise
            except Exception as err:
                error = merge(error, ValidationError([exception(err)]))
        else:
            break
    if error is not None:
        raise error
    return _obj


V = TypeVar("V", bound=Callable)


@overload
def validator(func: V) -> V:
    ...


@overload
def validator(field: Any, *, discard: bool = True) -> Callable[[V], V]:
    ...


def validator(arg=None, field=None, *, discard=True):
    if isfunction(arg):
        return Validator(arg)
    if arg is None and field is None or arg is not None and field is not None:
        raise ValueError("Bad use of validator")
    field = field or arg
    if not isinstance(field, Field):
        raise TypeError("validator argument must be a field")

    def decorator(func):
        @wraps(func)
        def wrapper(self, **kwargs):
            has_error = False
            for error in func(self, **kwargs):
                has_error = True
                if isinstance(error, str):
                    yield field, error
                else:
                    path, msg = error
                    if isinstance(path, Iterable) and not isinstance(path, str):
                        yield (field, *path), msg
                    else:
                        yield (field, path), msg
            if has_error and discard:
                raise Discard(field)

        return validator(wrapper)

    return decorator


VALIDATORS_METADATA = f"{PREFIX}validators"


def field_validator(*validators: ValidatorFunc) -> Metadata:
    return DictWithUnion({VALIDATORS_METADATA: list(map(SimpleValidator, validators))})



from enum import Enum
from typing import Any, Iterable, Mapping, Tuple, Type, TypeVar

from dataclasses import is_dataclass

from apischema.conversion import Converter, OutputVisitorMixin
from apischema.dataclasses import Field, get_output_fields_raw
from apischema.fields import get_fields_set
from apischema.types import AnyType, PRIMITIVE_TYPES
from apischema.visitor import NOT_CUSTOM, Unsupported

BUILTIN_TYPES = [*PRIMITIVE_TYPES, Iterable, Mapping]


def dataclass_field_value(obj: Any, field: Field):
    assert is_dataclass(obj)
    value = getattr(obj, field.name)
    if field.output_converter is not None:
        value = field.output_converter(value)
    return value


T = TypeVar("T", bound=Any)


class ToData(OutputVisitorMixin[Any, Any]):
    def __init__(self, conversions: Mapping[Type, Type], exclude_unset: bool):
        super().__init__(conversions)
        self.exclude_unset = exclude_unset

    def _custom(self, cls: Type, custom: Tuple[Type, Converter], obj: Any) -> Any:
        _, converter = custom
        return self.visit(converter(obj))

    def visit(self, obj: T, cls: Type[T] = None) -> Any:
        cls = cls or type(obj)
        if cls in PRIMITIVE_TYPES:
            return obj
        if cls is dict:
            return {self.visit(key): self.visit(value) for key, value in obj.items()}
        if cls in {list, tuple, set, frozenset}:
            return [self.visit(elt) for elt in obj]
        custom = self.custom(cls, obj)
        if custom is not NOT_CUSTOM:
            return custom
        if is_dataclass(cls):
            fields, properties_fields = get_output_fields_raw(cls)
            if self.exclude_unset:
                fields_set = get_fields_set(obj)
                fields = [f for f in fields if f.name in fields_set]
            result = {
                field.alias: self.visit(dataclass_field_value(obj, field))
                for field in fields
            }
            for field in properties_fields:
                value = self.visit(dataclass_field_value(obj, field))
                result.update(value)
            return result
        if issubclass(cls, Enum):
            return obj.value
        for builtin in BUILTIN_TYPES:
            if issubclass(cls, builtin):
                return self.visit(obj, builtin)
        raise Unsupported(cls)


def to_data(
    obj: Any, *, conversions: Mapping[Type, AnyType] = None, exclude_unset: bool = True,
) -> Any:
    return ToData(conversions or {}, exclude_unset).visit(obj)
